"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var R = require("ramda");
var PactNative = require("../native/index.node");
var MatchersV3;
(function (MatchersV3) {
    /**
     * Value must match the given template
     * @param template Template to base the comparison on
     */
    function like(template) {
        return {
            "pact:matcher:type": "type",
            value: template,
        };
    }
    MatchersV3.like = like;
    /**
     * Array where each element must match the given template
     * @param template Template to base the comparison on
     */
    function eachLike(template) {
        return {
            "pact:matcher:type": "type",
            value: [template],
        };
    }
    MatchersV3.eachLike = eachLike;
    /**
     * An array that has to have at least one element and each element must match the given template
     * @param template Template to base the comparison on
     * @param count Number of examples to generate, defaults to one
     */
    function atLeastOneLike(template, count) {
        if (count === void 0) { count = 1; }
        return {
            min: 1,
            "pact:matcher:type": "type",
            value: R.times(function () { return template; }, count),
        };
    }
    MatchersV3.atLeastOneLike = atLeastOneLike;
    /**
     * An array that has to have at least the required number of elements and each element must match the given template
     * @param template Template to base the comparison on
     * @param min Minimum number of elements required in the array
     * @param count Number of examples to generate, defaults to one
     */
    function atLeastLike(template, min, count) {
        var elements = count || min;
        if (count && count < min) {
            throw new Error("atLeastLike has a minimum of " +
                min +
                " but " +
                count +
                " elements where requested." +
                " Make sure the count is greater than or equal to the min.");
        }
        return {
            min: min,
            "pact:matcher:type": "type",
            value: R.times(function () { return template; }, elements),
        };
    }
    MatchersV3.atLeastLike = atLeastLike;
    /**
     * An array that has to have at most the required number of elements and each element must match the given template
     * @param template Template to base the comparison on
     * @param max Maximum number of elements required in the array
     * @param count Number of examples to generate, defaults to one
     */
    function atMostLike(template, max, count) {
        var elements = count || 1;
        if (count && count > max) {
            throw new Error("atMostLike has a maximum of " +
                max +
                " but " +
                count +
                " elements where requested." +
                " Make sure the count is less than or equal to the max.");
        }
        return {
            max: max,
            "pact:matcher:type": "type",
            value: R.times(function () { return template; }, elements),
        };
    }
    MatchersV3.atMostLike = atMostLike;
    /**
     * An array whose size is constrained to the minimum and maximum number of elements and each element must match the given template
     * @param template Template to base the comparison on
     * @param min Minimum number of elements required in the array
     * @param max Maximum number of elements required in the array
     * @param count Number of examples to generate, defaults to one
     */
    function constrainedArrayLike(template, min, max, count) {
        var elements = count || min;
        if (count) {
            if (count < min) {
                throw new Error("constrainedArrayLike has a minimum of " +
                    min +
                    " but " +
                    count +
                    " elements where requested." +
                    " Make sure the count is greater than or equal to the min.");
            }
            else if (count > max) {
                throw new Error("constrainedArrayLike has a maximum of " +
                    max +
                    " but " +
                    count +
                    " elements where requested." +
                    " Make sure the count is less than or equal to the max.");
            }
        }
        return {
            min: min,
            max: max,
            "pact:matcher:type": "type",
            value: R.times(function () { return template; }, elements),
        };
    }
    MatchersV3.constrainedArrayLike = constrainedArrayLike;
    /**
     * Value must be a boolean
     * @param b Boolean example value
     */
    function boolean(b) {
        return {
            "pact:matcher:type": "type",
            value: b,
        };
    }
    MatchersV3.boolean = boolean;
    /**
     * Value must be an integer (must be a number and have no decimal places)
     * @param int Example value. If omitted a random value will be generated.
     */
    function integer(int) {
        if (int) {
            return {
                "pact:matcher:type": "integer",
                value: int,
            };
        }
        else {
            return {
                "pact:generator:type": "RandomInt",
                "pact:matcher:type": "integer",
                value: 101,
            };
        }
    }
    MatchersV3.integer = integer;
    /**
     * Value must be a decimal number (must be a number and have decimal places)
     * @param num Example value. If omitted a random value will be generated.
     */
    function decimal(num) {
        if (num) {
            return {
                "pact:matcher:type": "decimal",
                value: num,
            };
        }
        else {
            return {
                "pact:generator:type": "RandomDecimal",
                "pact:matcher:type": "decimal",
                value: 12.34,
            };
        }
    }
    MatchersV3.decimal = decimal;
    /**
     * Value must be a number
     * @param num Example value. If omitted a random integer value will be generated.
     */
    function number(num) {
        if (num) {
            return {
                "pact:matcher:type": "number",
                value: num,
            };
        }
        else {
            return {
                "pact:generator:type": "RandomInt",
                "pact:matcher:type": "number",
                value: 1234,
            };
        }
    }
    MatchersV3.number = number;
    /**
     * Value must be a string
     * @param str Example value
     */
    function string(str) {
        return {
            "pact:matcher:type": "type",
            value: str,
        };
    }
    MatchersV3.string = string;
    function regex(pattern, str) {
        if (pattern instanceof RegExp) {
            return {
                "pact:matcher:type": "regex",
                regex: pattern.source,
                value: str,
            };
        }
        return {
            "pact:matcher:type": "regex",
            regex: pattern,
            value: str,
        };
    }
    MatchersV3.regex = regex;
    /**
     * Value that must be equal to the example. This is mainly used to reset the matching rules which cascade.
     * @param value Example value
     */
    function equal(value) {
        return {
            "pact:matcher:type": "equality",
            value: value,
        };
    }
    MatchersV3.equal = equal;
    /**
     * String value that must match the provided datetime format string.
     * @param format Datetime format string. See [Java SimpleDateFormat](https://docs.oracle.com/javase/8/docs/api/java/text/SimpleDateFormat.html)
     * @param example Example value to use. If omitted a value using the current system date and time will be generated.
     */
    function timestamp(format, example) {
        return datetime(format, example);
    }
    MatchersV3.timestamp = timestamp;
    /**
     * String value that must match the provided datetime format string.
     * @param format Datetime format string. See [Java SimpleDateFormat](https://docs.oracle.com/javase/8/docs/api/java/text/SimpleDateFormat.html)
     * @param example Example value to use. If omitted a value using the current system date and time will be generated.
     */
    function datetime(format, example) {
        return {
            "pact:generator:type": "DateTime",
            "pact:matcher:type": "timestamp",
            format: format,
            value: example || PactNative.generate_datetime_string(format),
        };
    }
    MatchersV3.datetime = datetime;
    /**
     * String value that must match the provided time format string.
     * @param format Time format string. See [Java SimpleDateFormat](https://docs.oracle.com/javase/8/docs/api/java/text/SimpleDateFormat.html)
     * @param example Example value to use. If omitted a value using the current system time will be generated.
     */
    function time(format, example) {
        return {
            "pact:generator:type": "Time",
            "pact:matcher:type": "time",
            format: format,
            value: example || PactNative.generate_datetime_string(format),
        };
    }
    MatchersV3.time = time;
    /**
     * String value that must match the provided date format string.
     * @param format Date format string. See [Java SimpleDateFormat](https://docs.oracle.com/javase/8/docs/api/java/text/SimpleDateFormat.html)
     * @param example Example value to use. If omitted a value using the current system date will be generated.
     */
    function date(format, example) {
        return {
            format: format,
            "pact:generator:type": "Date",
            "pact:matcher:type": "date",
            value: example || PactNative.generate_datetime_string(format),
        };
    }
    MatchersV3.date = date;
    /**
     * Value that must include the example value as a substring.
     * @param value String value to include
     */
    function includes(value) {
        return {
            "pact:matcher:type": "include",
            value: value,
        };
    }
    MatchersV3.includes = includes;
    /**
     * Value that must be null. This will only match the JSON Null value. For other content types, it will
     * match if the attribute is missing.
     */
    function nullValue() {
        return {
            "pact:matcher:type": "null",
        };
    }
    MatchersV3.nullValue = nullValue;
    /**
     * Matches a URL composed of a base path and a list of path fragments
     * @param basePath Base path of the URL
     * @param pathFragments list of path fragments, can be regular expressions
     */
    function url(basePath, pathFragments) {
        var regex = ".*";
        var example = basePath;
        for (var _i = 0, pathFragments_1 = pathFragments; _i < pathFragments_1.length; _i++) {
            var p = pathFragments_1[_i];
            if (p instanceof Object && p["pact:matcher:type"] == "regex") {
                regex += "\\/" + p["regex"];
                example += "/" + p["value"];
            }
            else if (p instanceof RegExp) {
                regex += "\\/" + p.source;
                example += "/" + PactNative.generate_regex_string(p.source);
            }
            else {
                regex += "\\/" + p.toString();
                example += "/" + p.toString();
            }
        }
        return {
            "pact:matcher:type": "regex",
            regex: regex + "$",
            value: example,
        };
    }
    MatchersV3.url = url;
    /**
     * Matches the items in an array against a number of variants. Matching is successful if each variant
     * occurs once in the array. Variants may be objects containing matching rules.
     */
    function arrayContaining() {
        var variants = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            variants[_i] = arguments[_i];
        }
        return {
            "pact:matcher:type": "arrayContains",
            variants: variants
        };
    }
    MatchersV3.arrayContaining = arrayContaining;
    /**
     * Marks a item as to be injected from the provider state
     * @param expression Expression to lookup in the provider state context
     * @param exampleValue Example value to use in the consumer test
     */
    function fromProviderState(expression, exampleValue) {
        return {
            "pact:matcher:type": "type",
            "pact:generator:type": "ProviderState",
            expression: expression,
            value: exampleValue
        };
    }
    MatchersV3.fromProviderState = fromProviderState;
})(MatchersV3 = exports.MatchersV3 || (exports.MatchersV3 = {}));
//# sourceMappingURL=matchers.js.map