"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var mockery = require("mockery");
var MockNative = {
    generate_datetime_string: function () { return ""; },
    generate_regex_string: function () { return ""; },
};
mockery.registerMock("../native/index.node", MockNative);
var chai = require("chai");
var matchers_1 = require("./matchers");
var expect = chai.expect;
describe("V3 Matchers", function () {
    describe("#like", function () {
        it("returns a JSON representation of a like matcher", function () {
            var result = matchers_1.MatchersV3.like({
                a: "b",
            });
            expect(result).to.deep.equal({
                "pact:matcher:type": "type",
                value: {
                    a: "b",
                },
            });
        });
    });
    describe("#eachLike", function () {
        it("returns a JSON representation of an eachLike matcher", function () {
            var result = matchers_1.MatchersV3.eachLike({
                a: "b",
            });
            expect(result).to.deep.equal({
                "pact:matcher:type": "type",
                value: [
                    {
                        a: "b",
                    },
                ],
            });
        });
    });
    describe("#atLeastOneLike", function () {
        describe("with no examples", function () {
            it("returns a JSON representation of an atLeastOneLike matcher", function () {
                var result = matchers_1.MatchersV3.atLeastOneLike({
                    a: "b",
                });
                expect(result).to.deep.equal({
                    "pact:matcher:type": "type",
                    min: 1,
                    value: [
                        {
                            a: "b",
                        },
                    ],
                });
            });
        });
        describe("when provided examples", function () {
            it("returns a JSON representation of an atLeastOneLike matcher with the correct number of examples", function () {
                var result = matchers_1.MatchersV3.atLeastOneLike({
                    a: "b",
                }, 4);
                expect(result).to.deep.equal({
                    "pact:matcher:type": "type",
                    min: 1,
                    value: [{ a: "b" }, { a: "b" }, { a: "b" }, { a: "b" }],
                });
            });
        });
    });
    describe("#atLeastLike", function () {
        describe("with no examples", function () {
            it("returns a JSON representation of an atLeastLike matcher", function () {
                var result = matchers_1.MatchersV3.atLeastLike({
                    a: "b",
                }, 2);
                expect(result).to.deep.equal({
                    "pact:matcher:type": "type",
                    min: 2,
                    value: [{ a: "b" }, { a: "b" }],
                });
            });
        });
        describe("when provided examples", function () {
            it("returns a JSON representation of an atLeastLike matcher with the correct number of examples", function () {
                var result = matchers_1.MatchersV3.atLeastLike({
                    a: "b",
                }, 2, 4);
                expect(result).to.deep.equal({
                    "pact:matcher:type": "type",
                    min: 2,
                    value: [{ a: "b" }, { a: "b" }, { a: "b" }, { a: "b" }],
                });
            });
        });
        it("throws an error if the number of examples is less than the minimum", function () {
            expect(function () { return matchers_1.MatchersV3.atLeastLike({ a: "b" }, 4, 2); }).to.throw("atLeastLike has a minimum of 4 but 2 elements where requested. Make sure the count is greater than or equal to the min.");
        });
    });
    describe("#atMostLike", function () {
        describe("with no examples", function () {
            it("returns a JSON representation of an atMostLike matcher", function () {
                var result = matchers_1.MatchersV3.atMostLike({
                    a: "b",
                }, 2);
                expect(result).to.deep.equal({
                    "pact:matcher:type": "type",
                    max: 2,
                    value: [{ a: "b" }],
                });
            });
        });
        describe("when provided examples", function () {
            it("returns a JSON representation of an atMostLike matcher with the correct number of examples", function () {
                var result = matchers_1.MatchersV3.atMostLike({
                    a: "b",
                }, 4, 4);
                expect(result).to.deep.equal({
                    "pact:matcher:type": "type",
                    max: 4,
                    value: [{ a: "b" }, { a: "b" }, { a: "b" }, { a: "b" }],
                });
            });
        });
        it("throws an error if the number of examples is more than the maximum", function () {
            expect(function () { return matchers_1.MatchersV3.atMostLike({ a: "b" }, 2, 4); }).to.throw("atMostLike has a maximum of 2 but 4 elements where requested. Make sure the count is less than or equal to the max.");
        });
    });
    describe("#constrainedArrayLike", function () {
        describe("with no examples", function () {
            it("returns a JSON representation of an constrainedArrayLike matcher", function () {
                var result = matchers_1.MatchersV3.constrainedArrayLike({
                    a: "b",
                }, 2, 4);
                expect(result).to.deep.equal({
                    "pact:matcher:type": "type",
                    min: 2,
                    max: 4,
                    value: [{ a: "b" }, { a: "b" }],
                });
            });
        });
        describe("when provided examples", function () {
            it("returns a JSON representation of an constrainedArrayLike matcher with the correct number of examples", function () {
                var result = matchers_1.MatchersV3.constrainedArrayLike({
                    a: "b",
                }, 2, 4, 3);
                expect(result).to.deep.equal({
                    "pact:matcher:type": "type",
                    min: 2,
                    max: 4,
                    value: [{ a: "b" }, { a: "b" }, { a: "b" }],
                });
            });
        });
        it("throws an error if the number of examples is less than the minimum", function () {
            expect(function () { return matchers_1.MatchersV3.constrainedArrayLike({ a: "b" }, 4, 6, 2); }).to.throw("constrainedArrayLike has a minimum of 4 but 2 elements where requested. Make sure the count is greater than or equal to the min.");
        });
        it("throws an error if the number of examples is more than the maximum", function () {
            expect(function () { return matchers_1.MatchersV3.constrainedArrayLike({ a: "b" }, 4, 6, 8); }).to.throw("constrainedArrayLike has a maximum of 6 but 8 elements where requested. Make sure the count is less than or equal to the max.");
        });
    });
    describe("#integer", function () {
        it("returns a JSON representation of an integer matcher", function () {
            var result = matchers_1.MatchersV3.integer(100);
            expect(result).to.deep.equal({
                "pact:matcher:type": "integer",
                value: 100,
            });
        });
        describe("when no example is given", function () {
            it("also includes a random integer generator", function () {
                var result = matchers_1.MatchersV3.integer();
                expect(result).to.deep.equal({
                    "pact:matcher:type": "integer",
                    "pact:generator:type": "RandomInt",
                    value: 101,
                });
            });
        });
    });
    describe("#decimal", function () {
        it("returns a JSON representation of an decimal matcher", function () {
            var result = matchers_1.MatchersV3.decimal(100.3);
            expect(result).to.deep.equal({
                "pact:matcher:type": "decimal",
                value: 100.3,
            });
        });
        describe("when no example is given", function () {
            it("also includes a random decimal generator", function () {
                var result = matchers_1.MatchersV3.decimal();
                expect(result).to.deep.equal({
                    "pact:matcher:type": "decimal",
                    "pact:generator:type": "RandomDecimal",
                    value: 12.34,
                });
            });
        });
    });
    describe("#number", function () {
        it("returns a JSON representation of an number matcher", function () {
            var result = matchers_1.MatchersV3.number(100.3);
            expect(result).to.deep.equal({
                "pact:matcher:type": "number",
                value: 100.3,
            });
        });
        describe("when no example is given", function () {
            it("also includes a random integer generator", function () {
                var result = matchers_1.MatchersV3.number();
                expect(result).to.deep.equal({
                    "pact:matcher:type": "number",
                    "pact:generator:type": "RandomInt",
                    value: 1234,
                });
            });
        });
    });
    describe("#boolean", function () {
        it("returns a JSON representation of a like matcher", function () {
            var result = matchers_1.MatchersV3.boolean(true);
            expect(result).to.deep.equal({
                "pact:matcher:type": "type",
                value: true,
            });
        });
    });
    describe("#string", function () {
        it("returns a JSON representation of a like matcher", function () {
            var result = matchers_1.MatchersV3.string("true");
            expect(result).to.deep.equal({
                "pact:matcher:type": "type",
                value: "true",
            });
        });
    });
    describe("#regex", function () {
        it("returns a JSON representation of a regex matcher", function () {
            var result = matchers_1.MatchersV3.regex("\\d+", "1234");
            expect(result).to.deep.equal({
                "pact:matcher:type": "regex",
                regex: "\\d+",
                value: "1234",
            });
        });
        describe("when given a regular expression", function () {
            it("returns a JSON representation of a regex matcher", function () {
                var result = matchers_1.MatchersV3.regex(/\d+/, "1234");
                expect(result).to.deep.equal({
                    "pact:matcher:type": "regex",
                    regex: "\\d+",
                    value: "1234",
                });
            });
        });
    });
    describe("#equal", function () {
        it("returns a JSON representation of an equality matcher", function () {
            var result = matchers_1.MatchersV3.equal("true");
            expect(result).to.deep.equal({
                "pact:matcher:type": "equality",
                value: "true",
            });
        });
    });
    describe("#datetime", function () {
        it("returns a JSON representation of a datetime matcher", function () {
            var result = matchers_1.MatchersV3.datetime("yyyy-MM-dd'T'HH:mm:ss.SSSX", "2016-02-11T09:46:56.023Z");
            expect(result).to.deep.equal({
                "pact:generator:type": "DateTime",
                "pact:matcher:type": "timestamp",
                format: "yyyy-MM-dd'T'HH:mm:ss.SSSX",
                value: "2016-02-11T09:46:56.023Z",
            });
        });
        describe("when no example is given", function () {
            it("generates a datetime from the current system time", function () {
                MockNative.generate_datetime_string = function () { return "2016-02-11T09:46:56.023Z"; };
                var result = matchers_1.MatchersV3.datetime("yyyy-MM-dd'T'HH:mm:ss.SSSX");
                expect(result).to.deep.equal({
                    "pact:generator:type": "DateTime",
                    "pact:matcher:type": "timestamp",
                    format: "yyyy-MM-dd'T'HH:mm:ss.SSSX",
                    value: "2016-02-11T09:46:56.023Z",
                });
            });
        });
    });
    describe("#time", function () {
        it("returns a JSON representation of a time matcher", function () {
            var result = matchers_1.MatchersV3.time("HH:mm:ss", "09:46:56");
            expect(result).to.deep.equal({
                "pact:generator:type": "Time",
                "pact:matcher:type": "time",
                format: "HH:mm:ss",
                value: "09:46:56",
            });
        });
        describe("when no example is given", function () {
            it("generates a time from the current system time", function () {
                MockNative.generate_datetime_string = function () { return "10:46:56.023"; };
                var result = matchers_1.MatchersV3.time("HH:mm:ss.SSS");
                expect(result).to.deep.equal({
                    "pact:generator:type": "Time",
                    "pact:matcher:type": "time",
                    format: "HH:mm:ss.SSS",
                    value: "10:46:56.023",
                });
            });
        });
    });
    describe("#date", function () {
        it("returns a JSON representation of a date matcher", function () {
            var result = matchers_1.MatchersV3.date("yyyy-MM-dd", "2016-02-11");
            expect(result).to.deep.equal({
                "pact:generator:type": "Date",
                "pact:matcher:type": "date",
                format: "yyyy-MM-dd",
                value: "2016-02-11",
            });
        });
        describe("when no example is given", function () {
            it("generates a date from the current system time", function () {
                MockNative.generate_datetime_string = function () { return "2020-02-11"; };
                var result = matchers_1.MatchersV3.date("yyyy-MM-dd");
                expect(result).to.deep.equal({
                    "pact:generator:type": "Date",
                    "pact:matcher:type": "date",
                    format: "yyyy-MM-dd",
                    value: "2020-02-11",
                });
            });
        });
    });
    describe("#includes", function () {
        it("returns a JSON representation of an include matcher", function () {
            var result = matchers_1.MatchersV3.includes("true");
            expect(result).to.deep.equal({
                "pact:matcher:type": "include",
                value: "true",
            });
        });
    });
    describe("#nullValue", function () {
        it("returns a JSON representation of an null matcher", function () {
            var result = matchers_1.MatchersV3.nullValue();
            expect(result).to.deep.equal({
                "pact:matcher:type": "null",
            });
        });
    });
    describe("#url", function () {
        it("returns a JSON representation of a regex matcher for the URL", function () {
            var result = matchers_1.MatchersV3.url("http://localhost:8080", [
                "users",
                "1234",
                "posts",
                "latest",
            ]);
            expect(result).to.deep.equal({
                "pact:matcher:type": "regex",
                regex: ".*\\/users\\/1234\\/posts\\/latest$",
                value: "http://localhost:8080/users/1234/posts/latest",
            });
        });
        describe("when provided with a regex matcher", function () {
            it("returns a JSON representation of a regex matcher for the URL", function () {
                var result = matchers_1.MatchersV3.url("http://localhost:8080", [
                    "users",
                    matchers_1.MatchersV3.regex("\\d+", "1234"),
                    "posts",
                    "latest",
                ]);
                expect(result).to.deep.equal({
                    "pact:matcher:type": "regex",
                    regex: ".*\\/users\\/\\d+\\/posts\\/latest$",
                    value: "http://localhost:8080/users/1234/posts/latest",
                });
            });
        });
        describe("when provided with a regular expression", function () {
            it("returns a JSON representation of a regex matcher for the URL", function () {
                MockNative.generate_regex_string = function () { return "12345678"; };
                var result = matchers_1.MatchersV3.url("http://localhost:8080", [
                    "users",
                    /\d+/,
                    "posts",
                    "latest",
                ]);
                expect(result).to.deep.equal({
                    "pact:matcher:type": "regex",
                    regex: ".*\\/users\\/\\d+\\/posts\\/latest$",
                    value: "http://localhost:8080/users/12345678/posts/latest",
                });
            });
        });
    });
});
//# sourceMappingURL=matchers.spec.js.map