"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ramda_1 = require("ramda");
var pkg = require("../common/metadata");
var PactNative = require("../native/index.node");
function displayQuery(query) {
    var pairs = ramda_1.toPairs(query);
    var mapped = ramda_1.flatten(ramda_1.map(function (_a) {
        var key = _a[0], values = _a[1];
        return ramda_1.map(function (val) { return key + "=" + val; }, values);
    }, pairs));
    return ramda_1.join('&', mapped);
}
function displayHeaders(headers, indent) {
    return ramda_1.join('\n' + indent, ramda_1.map(function (_a) {
        var k = _a[0], v = _a[1];
        return k + ": " + v;
    }, ramda_1.toPairs(headers)));
}
function displayRequest(request, indent) {
    var output = "\n" + indent + "Method: " + request.method + "\n" + indent + "Path: " + request.path;
    if (request.query) {
        output += "\n" + indent + "Query String: " + displayQuery(request.query);
    }
    if (request.headers) {
        output += "\n" + indent + "Headers:\n" + indent + "  " + displayHeaders(request.headers, indent + '  ');
    }
    if (request.body) {
        output += "\n" + indent + "Body: " + request.body.substr(0, 20) + "... (" + request.body.length + " length)";
    }
    return output;
}
function generateMockServerError(testResult, indent) {
    var error = "Mock server failed with the following mismatches: ";
    var i = 1;
    for (var _i = 0, _a = testResult.mockServerMismatches; _i < _a.length; _i++) {
        var mismatchJson = _a[_i];
        var mismatches = JSON.parse(mismatchJson);
        if (mismatches.mismatches) {
            for (var _b = 0, _c = mismatches.mismatches; _b < _c.length; _b++) {
                var mismatch = _c[_b];
                error += "\n" + indent + i++ + ") " + mismatch.type + " " + (mismatch.path ? "(at " + mismatch.path + ") " : "") + mismatch.mismatch;
            }
        }
        if (mismatches.type == "request-not-found") {
            error += "\n\n" + indent + i++ + ") The following request was not expected: " + displayRequest(mismatches.request, indent + "    ");
        }
        if (mismatches.type == "missing-request") {
            error += "\n\n" + indent + i++ + ") The following request was expected but not received: " + displayRequest(mismatches.request, indent + "    ");
        }
    }
    return error;
}
var PactV3 = /** @class */ (function () {
    function PactV3(opts) {
        this.states = [];
        this.opts = opts;
        this.pact = new PactNative.Pact(opts.consumer, opts.provider, pkg.version, ramda_1.omit(['consumer', 'provider', 'dir'], opts));
    }
    PactV3.prototype.given = function (providerState, parameters) {
        this.states.push({ description: providerState, parameters: parameters });
        return this;
    };
    PactV3.prototype.uponReceiving = function (desc) {
        this.pact.addInteraction(desc, this.states);
        return this;
    };
    PactV3.prototype.withRequest = function (req) {
        this.pact.addRequest(req, req.body && JSON.stringify(req.body));
        return this;
    };
    PactV3.prototype.withRequestBinaryFile = function (req, contentType, file) {
        this.pact.addRequestBinaryFile(req, contentType, file);
        return this;
    };
    PactV3.prototype.withRequestMultipartFileUpload = function (req, contentType, file, part) {
        this.pact.addRequestMultipartFileUpload(req, contentType, file, part);
        return this;
    };
    PactV3.prototype.willRespondWith = function (res) {
        this.pact.addResponse(res, res.body && JSON.stringify(res.body));
        this.states = [];
        return this;
    };
    PactV3.prototype.withResponseBinaryFile = function (res, contentType, file) {
        this.pact.addResponseBinaryFile(res, contentType, file);
        return this;
    };
    PactV3.prototype.withResponseMultipartFileUpload = function (req, contentType, file, part) {
        this.pact.addResponseMultipartFileUpload(req, contentType, file, part);
        return this;
    };
    PactV3.prototype.executeTest = function (testFn) {
        var _this = this;
        var result = this.pact.executeTest(testFn, this.opts);
        if (result.testResult) {
            return result.testResult
                .then(function (val) {
                var testResult = _this.pact.getTestResult(result.mockServer.id);
                if (testResult.mockServerError) {
                    return Promise.reject(new Error(testResult.mockServerError));
                }
                else if (testResult.mockServerMismatches) {
                    return Promise.reject(new Error(generateMockServerError(testResult, "  ")));
                }
                else {
                    _this.pact.writePactFile(result.mockServer.id, _this.opts.dir);
                    return val;
                }
            })
                .catch(function (err) {
                var testResult = _this.pact.getTestResult(result.mockServer.id);
                if (testResult.mockServerError || testResult.mockServerMismatches) {
                    var error = "Test failed for the following reasons:";
                    error += "\n\n  Test code failed with an error: " + err.message;
                    if (err.stack) {
                        error += "\n" + err.stack + "\n";
                    }
                    if (testResult.mockServerError) {
                        error += "\n\n  " + testResult.mockServerError;
                    }
                    if (testResult.mockServerMismatches) {
                        error += "\n\n  " + generateMockServerError(testResult, "    ");
                    }
                    return Promise.reject(new Error(error));
                }
                else {
                    return Promise.reject(err);
                }
            })
                .finally(function () {
                _this.pact.shutdownTest(result);
            });
        }
        else {
            this.pact.shutdownTest(result);
            return Promise.reject(result.testError);
        }
    };
    return PactV3;
}());
exports.PactV3 = PactV3;
//# sourceMappingURL=pact.js.map